from collections import deque
from .operators import apply
from .contexts import PathContext


class LooseDictWalker(object):
    context_factory = PathContext

    def __init__(self, on_container=None, on_data=None, context_factory=None):
        self.on_container = on_container
        self.on_data = on_data
        self.context_factory = context_factory or self.__class__.context_factory

    def on_found(self, ctx, d, k):
        if self.on_container is not None:
            ctx(self, self.on_container, d)
        if self.on_data is not None:
            ctx(self, self.on_data, d[k])

    def create_context(self, ctx=None):
        return ctx or self.context_factory()

    def walk(self, qs, d, depth=-1, ctx=None):
        ctx = self.create_context(ctx)
        return self._walk(ctx, deque(qs), d, depth=depth)

    def _walk(self, ctx, qs, d, depth):
        if depth == 0:
            return

        if not qs:
            return

        if hasattr(d, "keys"):
            for k in list(d.keys()):
                ctx.push(k)
                if apply(qs[0], k):
                    q = qs.popleft()
                    self._walk(ctx, qs, d[k], depth - 1)
                    if len(qs) == 0:
                        self.on_found(ctx, d, k)
                    qs.appendleft(q)
                else:
                    self._walk(ctx, qs, d[k], depth)
                ctx.pop()
            return
        elif isinstance(d, (list, tuple)):
            ctx.push("[]")
            for e in d:
                self._walk(ctx, qs, e, depth)
            ctx.pop()
            return
        else:
            return



import re


def apply(q, v):
    if callable(q):
        return q(v)
    else:
        return q == v


def repr(self):
    return '<{self.__class__.__name__} args={args!r}>'.format(self=self, args=self.args)


class Regexp(object):
    __repr__ = repr

    def __init__(self, rx):
        if isinstance(rx, (str, bytes)):
            rx = re.compile(rx)
        self.args = rx

    def __call__(self, v):
        return self.args.search(v)


class Any(object):
    def __repr__(self):
        return '<{self.__class__.__name__}>'.format(self=self)

    def __call__(self, v):
        return True
ANY = Any()


class Not(object):
    __repr__ = repr

    def __init__(self, value):
        self.args = value

    def __call__(self, v):
        return not apply(self.args, v)


class Or(object):
    __repr__ = repr

    def __init__(self, args):
        self.args = args

    def __call__(self, v):
        for e in self.args:
            if apply(e, v):
                return True
        return False


class And(object):
    __repr__ = repr

    def __init__(self, args):
        self.args = args

    def __call__(self, v):
        for e in self.args:
            if not apply(e, v):
                return False
        return True



# with walkers


class SimpleContext(object):
    def push(self, ctx):
        pass

    def pop(self):
        pass

    def new_child(self):
        return self.__class__()

    def __call__(self, walker, fn, value):
        return fn(value)


class PathContext(object):
    def __init__(self):
        self.path = []

    def push(self, v):
        self.path.append(v)

    def pop(self):
        self.path.pop()


    def __call__(self, walker, fn, value):
        return fn(self.path, value)


class RecPathContext(PathContext):
    def __call__(self, walker, fn, value):
        return fn(walker, self.path, value)



# -*- coding:utf-8 -*-


def deepequal(d0, d1):
    return halfequal(d0, d1) and halfequal(d1, d0)


def halfequal(left, right):
    if hasattr(left, "keys"):
        for k in left.keys():
            if k not in right:
                return False
            if not halfequal(left[k], right[k]):
                return False
        return True
    elif isinstance(left, (list, tuple)):
        if len(left) != len(right):
            return False
        for x, y in zip(left, right):
            if not halfequal(x, y):
                return False
        return True
    else:
        return left == right



from .walkers import LooseDictWalker
from .operators import ANY


class ChainedContext(object):
    def __init__(self, query_list, path=None, value=None, parent=None):
        self.query_list = query_list
        self.path = path or []
        self.value = value
        self.parent = parent

    def push(self, v):
        self.path.append(v)

    def pop(self):
        self.path.pop()

    def new_child(self, query_list=None, path=None, value=None):
        query_list = query_list or self.query_list
        path = path or self.path[:-1]
        return self.__class__(query_list, path=path, value=value, parent=self)

    @property
    def on_container(self):
        if not self.query_list:
            return None
        return self.query_list[0].on_container

    @property
    def on_data(self):
        if not self.query_list:
            return None
        return self.query_list[0].on_data

    def __getitem__(self, i):
        if i <= 0:
            return self
        else:
            ctx = self
            for _ in range(i):
                ctx = ctx.parent
            return

    def __call__(self, walker, fn, value):
        return fn(self, walker, value)


class ChainQuery(object):
    def __init__(self, source, qs, cont=None, on_container=None, on_data=None):
        self.on_container = on_container
        self.on_data = on_data
        self.source = source
        self.qs = qs
        self.cont = cont

    def __repr__(self):
        return "<ChainQury qs={!r}>".format(self.qs)

    def chain(self, qs, on_data=None, on_container=None):
        # constructing query but this is reversed.
        return self.__class__(self.source, qs, cont=self, on_container=on_container, on_data=on_data)

    def flatten(self, r):
        if self.cont is None:
            r.append(self)
            return r
        self.cont.flatten(r)
        r.append(self)
        return r

    def walk(self, d, on_container=None, on_data=None, **kwargs):
        def on_match(ctx, walker, value):
            if ctx.path:
                hook = ctx.parent.on_container or on_container
                if hook is not None:
                    if hook(ctx, walker, value) is False:
                        return
                hook = ctx.parent.on_data or on_data
                if hook is not None:
                    if hook(ctx, walker, value[ctx.path[-1]]) is False:
                        return

            if not ctx.query_list:
                return

            current_query, *rest = ctx.query_list
            new_ctx = ctx.new_child(rest, value=value)
            if callable(current_query.qs):
                current_query.qs(new_ctx, walker, value)
            else:
                qs = [ANY, *current_query.qs]
                walker.walk(qs, value, ctx=new_ctx)

        query_list = self.flatten([])
        ctx = self.source.context_factory(query_list, **kwargs)
        walker = self.source.walker_factory(on_container=on_match)
        return on_match(ctx, walker, d)


class ChainSource(object):
    def __init__(self, query_factory=ChainQuery, walker_factory=LooseDictWalker, context_factory=ChainedContext):
        self.query_factory = query_factory
        self.walker_factory = walker_factory
        self.context_factory = context_factory

    def chain(self, qs, on_container=None, on_data=None):
        return self.query_factory(self, qs, on_container=on_container, on_data=on_data)


chain = ChainSource



from dictknife.walkers import LooseDictWalker  # NOQA
from dictknife.chain import ChainSource, chain  # NOQA
from dictknife.accessor import Accessor  # NOQA
from dictknife.deepmerge import deepmerge  # NOQA
from dictknife.deepequal import deepequal  # NOQA
from dictknife.operators import (  # NOQA
    apply,
    Regexp,
    Any,
    Not,
    Or,
    And,
)



class Accessor(object):
    def __init__(self, make_dict=dict):
        self.make_dict = make_dict

    def assign(self, d, path, value):
        for name in path[:-1]:
            if name not in d:
                d[name] = self.make_dict()
            d = d[name]
        d[path[-1]] = value

    def access(self, d, path):
        for name in path:
            d = d[name]
        return d

    def maybe_remove(self, d, path):
        container = self.maybe_access_container(d, path)
        if container is not None:
            container.pop(path[-1])

    def exists(self, d, path):
        return self.maybe_access_container(d, path) is not None

    def maybe_access_container(self, d, path):
        for name in path[:-1]:
            if name not in d:
                return None
            d = d[name]
        if path[-1] not in d:
            return None
        return d



# -*- coding:utf-8 -*-
import functools


def _deepmerge(left, right):
    if isinstance(left, list):
        r = left[:]
        if isinstance(right, (list, tuple)):
            for e in right:
                if e not in r:
                    r.append(e)
        else:
            if right not in r:
                r.append(right)
        return r
    elif hasattr(left, "get"):
        if hasattr(right, "get"):
            r = left.copy()
            for k in right.keys():
                if k in left:
                    r[k] = _deepmerge(r[k], right[k])
                else:
                    r[k] = right[k]
            return r
        elif right is None:
            return left
        else:
            raise ValueError("cannot merge dict and non-dict: left=%s, right=%s", left, right)
    else:
        return right


def deepmerge(*ds):
    return functools.reduce(_deepmerge, ds)



# -*- coding:utf-8 -*-
import json
import yaml
import os.path
import logging
from collections import OrderedDict
logger = logging.getLogger(__name__)


class Format:
    yaml = "yaml"
    json = "json"
    unknown = "(unknown)"


def dispatch_by_format(filename, fn_map, default=Format.unknown):
    _, ext = os.path.splitext(filename)
    if ext in (".yaml", ".yml"):
        return fn_map[Format.yaml]
    elif ext in (".json", ".js"):
        return fn_map[Format.json]
    else:
        return fn_map[default]


def _json_load(fp):
    return json.load(fp, object_pairs_hook=OrderedDict)


def _json_dump(d, fp):
    return json.dump(d, fp, ensure_ascii=False, indent=2)


def _yaml_dump(d, fp):
    return yaml.dump(d, fp, allow_unicode=True, default_flow_style=False)


def load(fp, format=None, fn_map={Format.yaml: yaml.load, Format.json: _json_load, Format.unknown: yaml.load}):
    if format is not None:
        loader = fn_map[format]
    else:
        fname = getattr(fp, "name", "(unknown)")
        loader = dispatch_by_format(fname, fn_map, default=loading_config.input_format)
    return loader(fp)


def dump(d, fp, format=None, fn_map={Format.yaml: _yaml_dump, Format.json: _json_dump, Format.unknown: _yaml_dump}):
    if format is not None:
        dumper = fn_map[format]
    else:
        fname = getattr(fp, "name", "(unknown)")
        dumper = dispatch_by_format(fname, fn_map, default=loading_config.output_format)
    return dumper(d, fp)


class loading_config:
    input_format = Format.yaml
    output_format = Format.yaml


def setup(input=None, output=None):
    global loading_config
    if input is not None:
        logger.debug("setup input format: %s", input)
        loading_config.input_format = input
    if output is not None:
        logger.debug("setup output format: %s", output)
        loading_config.output_format = output
    yaml.add_constructor('tag:yaml.org,2002:map', construct_odict)
    yaml.add_representer(OrderedDict, represent_odict)


def represent_odict(dumper, instance):
    return dumper.represent_mapping('tag:yaml.org,2002:map', instance.items())


def construct_odict(loader, node):
    return OrderedDict(loader.construct_pairs(node))
